/*! ********************************************
 * SineGenerator.as
 * Copyright 2022 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> license.txt
 ******************************************** */
package {
import sora.graybox.GrayBox;
import sora.graybox.IGrayBoxModule;
import sora.graybox.MidiEvent;
import sora.graybox.ProcessData;
// サイン波ジェネレーターモジュール
public class SineGenerator implements IGrayBoxModule {
	public var volume:Number = 0.1;  // 音量
	private var _keyOn:Boolean = false;  // 出音フラグ
	private var _sampleRate:Number = 44100.0;  // サンプリング周波数
	private var _frequency:Number = 440.0;  // 出音周波数
	private var _af:Number = _frequency * 6.28318530718;  // 角周波数
	private var _phase:int = 0;  // 位相 (周期 = サンプリング周波数)
	// コンストラクター
	public function SineGenerator():void {}
	
// sora.graybox.IGrayBoxModule -->
	// ホストのサンプリングレートが変更された時に呼ばれるメソッド
	// GrayBox.setSampleRate() -> このメソッド
	public function updateSampleRate(sampleRate:Number):void {
		var n:Number = _phase / _sampleRate;
		n = int(n) - n;
		_sampleRate = sampleRate;
		_phase = Math.round(n * sampleRate);
	}
	// ホストのオーディオチャンネル数が変更された時に呼ばれるメソッド
	// GrayBox.setAudioChannels() -> このメソッド
	public function updateAudioChannels(ch:int):void{}
	// ホストのプロセスブロックサイズが変更された時に呼ばれるメソッド
	// GrayBox.setProcessBlockSize() -> このメソッド
	public function updateProcessBlockSize(size:int):void{}
	// ホストのGrayBox.resetModules()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.resetModules() -> このメソッド
	public function resetModule():void{ volume = 0.1; _keyOn = false; }
	// ホストのGrayBox.initiModules()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.initiModules() -> このメソッド
	public function initiModule():void{ volume = 0.1; _keyOn = false; }
	// このモジュールがホストへ接続された時に呼ばれるメソッド
	// GrayBox.attachModule() -> このメソッド
	public function connectModule(host:GrayBox):void {
		_keyOn = false;
		updateSampleRate(host.getSampleRate());
	}
	// このモジュールがホストから取り外された時に呼ばれるメソッド
	// GrayBox.detachModule() -> このメソッド
	public function disconnectModule():void{}
	// このモジュールが接続されているスロットの、バイパス状況が更新された時に呼ばれるメソッド
	// GrayBox.bypassProcess() -> このメソッド
	public function updateBypassStatus(bypass:Boolean):void{ if (bypass)_keyOn = false; }
	// ホストのGrayBox.startProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.startProcess() -> このメソッド
	public function startProcess():void{ _keyOn = false; }
	// ホストのGrayBox.stopProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.stopProcess() -> このメソッド
	public function stopProcess():void{ _keyOn = false; }
	// ホストのGrayBox.suspendProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.suspendProcess() -> このメソッド
	public function suspendProcess():void{}
	// ホストのGrayBox.resumeProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.resumeProcess() -> このメソッド
	public function resumeProcess():void{}
	// ホストが音声処理を実行した時に呼ばれるメソッド
	// GrayBox.process() -> このメソッド
	public function process(procData:ProcessData):void {
	// 音声処理準備 -->
		// ProcessDataオブジェクトはホストから音声処理に必要な情報を受け取り、
		// 各モジュールへ伝えるオブジェクトです。
		// ここではホストから渡されたオーディオバッファを参照して音声処理を行います。
		var _startFrame:int = procData.offsetFrames,  // 音声処理開始フレーム
			_endFrame:int = _startFrame + procData.sampleFrames;  // 終了フレーム
	// 音声処理、MIDIイベント処理 -->
		// GrayBoxシステムにおけるMIDIメッセージの伝送はProcessDataオブジェクトが持つ
		// MIDIイベントリスト(片方向リンクリスト)を使用して行われます。
		// MIDIメッセージの送信側はこのリストにMIDIイベントを登録していき、
		// MIDIメッセージの受信側はこのリストからMIDIイベント参照してメッセージの処理を行います。
		
		// ・ProcessData.midiEventList
		//   MIDIイベントリスト(先頭ノード(リスト管理用ダミーノード))
		//   このリストのノードの並びはMidiEvent.offsetFramesの昇順です。
		// ・ProcessData.getMidiEvents() : MidiEvent
		//   リスト先頭のダミーノードを除くリストを取得(通常はこちらを使用してください)
		//   リストに何も登録されていなければnullが返ります。
		var _event:MidiEvent = procData.getMidiEvents();
		// リストからMIDIイベントを順に読み込み、処理を行う
		while (_event != null) {
			// 出音フラグが立っていれば、このMIDIイベントまで音声処理
			// MidiEvent.offsetFrames -> ブロック先頭からのフレームオフセット
			if (_keyOn)_sineGen(procData.audioOut, _startFrame, _event.offsetFrames);
			
			// MIDIメッセージタイプ別で処理の振り分け
			switch (_event.type) {
				case 0x80: // ノートオフ
					_startFrame = _event.offsetFrames;  // 開始フレーム更新
					_keyOn = false;  // 出音フラグ オン
					break;
					
				case 0x90: // ノートオン
					_startFrame = _event.offsetFrames; // 開始フレーム更新
					_keyOn = true;  // 出音フラグ オン
					// 出音条件の更新
					_frequency = 440.0 * Math.pow(2.0, (_event.value0 - 69.0) / 12.0); // 周波数更新
					_af = _frequency * 6.28318530718; // 角周波数更新
					_phase = 0;  // 位相の初期化
					break;
			}
			// 次のMIDIイベント
			_event = _event.next;
		}
		// 終了フレームまで音声処理
		if (_keyOn)_sineGen(procData.audioOut, _startFrame, _endFrame);
	}
	private function _sineGen(out:Vector.<Vector.<Number>>, startFrame:int, frameEnd:int):void {
		var _outL:Vector.<Number> = out[0],
		    _outR:Vector.<Number> = out[1];
		for (var i:int = startFrame; i < frameEnd; i++) {
			var n:Number = Math.sin((_phase / _sampleRate) * _af) * volume; // サイン波出力
			_phase++;  // 位相の更新
			// オーディオバッファに対して加算代入
			_outL[i] += n;
			_outR[i] += n;
			// このモジュールしか音声処理を行わない場合は代入だけでも良い
			/* 
				_outL[i] = _n; 
				_outR[i] = _n;
			*/
		}
	}
}
}